#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "fileio.h"
#include "reverse.h"
#include "tiff.h"
#include "lzw.h"

/*

Tiff - Copyright 2005 Michael Kohn (mike@mikekohn.net)

This program is NOT opensourced.  You may not use any part
of this program for your own software.

*/

unsigned int read_long_offset(FILE *in, int offset)
{
int marker,r;

  marker=ftell(in);
  fseek(in,offset,SEEK_SET);

  r=read_long(in);

  fseek(in,marker,SEEK_SET);

  return r;
}

int read_tiff_header(FILE *in, struct tiff_header_t *tiff_header)
{
  tiff_header->byte_order[0]=getc(in);
  tiff_header->byte_order[1]=getc(in);

  if (tiff_header->byte_order[0]=='M')
  {
    tiff_header->version=read_word_b(in);
    tiff_header->pointer_to_ifd=read_long_b(in);
  }
    else
  {
    tiff_header->version=read_word(in);
    tiff_header->pointer_to_ifd=read_long(in);
  }

#ifdef DEBUG
  printf("--------------------------------------------\n");
  printf("TIFF HEADER\n");
  printf("--------------------------------------------\n");
  printf("          Byte Order: %c%c ",tiff_header->byte_order[0],tiff_header->byte_order[1]);
  if (tiff_header->byte_order[0]=='M') printf(" (BIG ENDIAN)\n");
    else
  if (tiff_header->byte_order[0]=='I') printf(" (LITTLE ENDIAN)\n");
  printf("             Version: %d\n",tiff_header->version);
  printf("Pointer To First IFD: %d\n",tiff_header->pointer_to_ifd);
#endif

  return 0;
}

int read_tiff_ifd(FILE *in, struct tiff_ifd_t *tiff_ifd, int endian)
{
#ifdef DEBUG
int t,ch,marker;
#endif

  if (endian==1)
  {
    tiff_ifd->tag_type=read_word_b(in);
    tiff_ifd->data_type=read_word_b(in);
    tiff_ifd->length=read_long_b(in);
    tiff_ifd->value=read_long_b(in);

    if (tiff_ifd->data_type==3 || tiff_ifd->data_type==8)
    { tiff_ifd->value>>=16; }
      else
    if (tiff_ifd->data_type==1 || tiff_ifd->data_type==6 ||
        tiff_ifd->data_type==7)
    { tiff_ifd->value>>=24; }
  }
    else
  {
    tiff_ifd->tag_type=read_word(in);
    tiff_ifd->data_type=read_word(in);
    tiff_ifd->length=read_long(in);
    tiff_ifd->value=read_long(in);
  }

#ifdef DEBUG
  printf("--------------------------------------------\n");
  printf("TIFF IFD\n");
  printf("--------------------------------------------\n");
  printf("            Tag Type: 0x%04x ",tiff_ifd->tag_type);

  if (tiff_ifd->tag_type==0xfe) printf("(NewSubfile)\n");
    else
  if (tiff_ifd->tag_type==0xff) printf("(SubfileType)\n");
    else
  if (tiff_ifd->tag_type==0x100) printf("(ImageWidth)\n");
    else
  if (tiff_ifd->tag_type==0x101) printf("(ImageLength)\n");
    else
  if (tiff_ifd->tag_type==0x112) printf("(Orientation)\n");
    else
  if (tiff_ifd->tag_type==0x11a) printf("(XResolution)\n");
    else
  if (tiff_ifd->tag_type==0x11b) printf("(YResolution)\n");
    else
  if (tiff_ifd->tag_type==0x11c) printf("(PlanarConfiguration)\n");
    else
  if (tiff_ifd->tag_type==0x128) printf("(ResolutionUnit)\n");
    else
  if (tiff_ifd->tag_type==0x111) printf("(StripOffsets)\n");
    else
  if (tiff_ifd->tag_type==0x117) printf("(StripByteCounts)\n");
    else
  if (tiff_ifd->tag_type==0x116) printf("(RowsPerStrip)\n");
    else
  if (tiff_ifd->tag_type==0x142) printf("(TileWidth)\n");
    else
  if (tiff_ifd->tag_type==0x143) printf("(TileLength)\n");
    else
  if (tiff_ifd->tag_type==0x144) printf("(TileOffsets)\n");
    else
  if (tiff_ifd->tag_type==0x145) printf("(TileByteCounts)\n");
    else
  if (tiff_ifd->tag_type==0x102) printf("(BitsPerSample)\n");
    else
  if (tiff_ifd->tag_type==0x106) printf("(Photometric Interpretation)\n");
    else
  if (tiff_ifd->tag_type==0x107) printf("(Thresholding)\n");
    else
  if (tiff_ifd->tag_type==0x108) printf("(CellWdith)\n");
    else
  if (tiff_ifd->tag_type==0x109) printf("(CellLength)\n");
    else
  if (tiff_ifd->tag_type==0x115) printf("(SamplesPerPixel)\n");
    else
  if (tiff_ifd->tag_type==0x118) printf("(MinSamplesValue)\n");
    else
  if (tiff_ifd->tag_type==0x119) printf("(MaxSamplesValue)\n");
    else
  if (tiff_ifd->tag_type==0x122) printf("(GrayResponseUnit)\n");
    else
  if (tiff_ifd->tag_type==0x123) printf("(GrayResponseCurve)\n");
    else
  if (tiff_ifd->tag_type==0x12c) printf("(ColorResponseUnit)\n");
    else
  if (tiff_ifd->tag_type==0x12d) printf("(ColorResponseCurve)\n");
    else
  if (tiff_ifd->tag_type==0x131) printf("(Software)\n");
    else
  if (tiff_ifd->tag_type==0x132) printf("(Date/Time)\n");
    else
  if (tiff_ifd->tag_type==0x13b) printf("(Artist)\n");
    else
  if (tiff_ifd->tag_type==0x13c) printf("(HostComputer)\n");
    else
  if (tiff_ifd->tag_type==0x13d) printf("(Predictor)\n");
    else
  if (tiff_ifd->tag_type==0x13e) printf("(WhitePoint)\n");
    else
  if (tiff_ifd->tag_type==0x13f) printf("(PrimaryChromatics)\n");
    else
  if (tiff_ifd->tag_type==0x140) printf("(ColorMap)\n");
    else
  if (tiff_ifd->tag_type==0x10a) printf("(FillOrder)\n");
    else
  if (tiff_ifd->tag_type==0x103) printf("(Compression)\n");
    else
  if (tiff_ifd->tag_type==0x124) printf("(T4Options)\n");
    else
  if (tiff_ifd->tag_type==0x125) printf("(T6Options)\n");
    else
  if (tiff_ifd->tag_type==0x152) printf("(ExtraSamples)\n");
    else
  if (tiff_ifd->tag_type==0x153) printf("(SampleFormat)\n");
    else
  if (tiff_ifd->tag_type==0x154) printf("(SMinSampleValue)\n");
    else
  if (tiff_ifd->tag_type==0x155) printf("(SMaxSampleValue)\n");
    else
  if (tiff_ifd->tag_type==0x156) printf("(TransferRange)\n");
    else
  if (tiff_ifd->tag_type==0x10d) printf("(DocumentName)\n");
    else
  if (tiff_ifd->tag_type==0x10e) printf("(ImageDescription)\n");
    else
  if (tiff_ifd->tag_type==0x10f) printf("(ScannerMake)\n");
    else
  if (tiff_ifd->tag_type==0x110) printf("(ScannerModel)\n");
    else
  if (tiff_ifd->tag_type==0x11d) printf("(PageName)\n");
    else
  if (tiff_ifd->tag_type==0x11e) printf("(XPosition)\n");
    else
  if (tiff_ifd->tag_type==0x11f) printf("(YPosition)\n");
    else
  if (tiff_ifd->tag_type==0x129) printf("(PageNumber)\n");
    else
  if (tiff_ifd->tag_type==0x8298) printf("(Copyright)\n");
    else
  if (tiff_ifd->tag_type==0x120) printf("(FreeOffsets)\n");
    else
  if (tiff_ifd->tag_type==0x121) printf("(FreeByteCounts)\n");
    else
  if (tiff_ifd->tag_type==0x14c) printf("(InkSet)\n");
    else
  if (tiff_ifd->tag_type==0x14d) printf("(InkNames)\n");
    else
  if (tiff_ifd->tag_type==0x14e) printf("(NumberOfInks)\n");
    else
  if (tiff_ifd->tag_type==0x150) printf("(DotRange)\n");
    else
  if (tiff_ifd->tag_type==0x151) printf("(TargetPrinter)\n");
    else
  if (tiff_ifd->tag_type==0x200) printf("(JPEGProc)\n");
    else
  if (tiff_ifd->tag_type==0x201) printf("(JPEGInterchangeFormat)\n");
    else
  if (tiff_ifd->tag_type==0x203) printf("(JPEGInterchangeFormatLength)\n");
    else
  if (tiff_ifd->tag_type==0x204) printf("(JPEGRestartInterval)\n");
    else
  if (tiff_ifd->tag_type==0x205) printf("(JPEGLossLessPredictors)\n");
    else
  if (tiff_ifd->tag_type==0x206) printf("(JpegPointTransforms)\n");
    else
  if (tiff_ifd->tag_type==0x207) printf("(JpegQTables)\n");
    else
  if (tiff_ifd->tag_type==0x208) printf("(JpegDCTables)\n");
    else
  if (tiff_ifd->tag_type==0x209) printf("(JpegACTables)\n");
    else
  if (tiff_ifd->tag_type==0x211) printf("(YCbCrCoefficients)\n");
    else
  if (tiff_ifd->tag_type==0x212) printf("(YCbCrSubSampling)\n");
    else
  if (tiff_ifd->tag_type==0x213) printf("(YCbCrPositioning)\n");
    else
  { printf("(Not Known)\n"); }

  printf("           Data Type: %d ",tiff_ifd->data_type);

  if (tiff_ifd->data_type==0x01) printf("(Byte [8 bit])\n");
    else
  if (tiff_ifd->data_type==0x02) printf("(ASCII CODE [8 bit])\n");
    else
  if (tiff_ifd->data_type==0x03) printf("(SHORT [16 bit unsigned])\n");
    else
  if (tiff_ifd->data_type==0x04) printf("(LONG [32 bit unsigned])\n");
    else
  if (tiff_ifd->data_type==0x05) printf("(RATIONAL [2 longs])\n");
    else
  if (tiff_ifd->data_type==0x06) printf("(SBYTE [8 bit signed])\n");
    else
  if (tiff_ifd->data_type==0x07) printf("(UNDEFINED [8 bit anything])\n");
    else
  if (tiff_ifd->data_type==0x08) printf("(SSHORT [16 bit signed])\n");
    else
  if (tiff_ifd->data_type==0x09) printf("(SLONG [32 bit signed])\n");
    else
  if (tiff_ifd->data_type==0x0a) printf("(RATIONAL [2 SLONG])\n");
    else
  if (tiff_ifd->data_type==0x0b) printf("(FLOAT [4 byte single])\n");
    else
  if (tiff_ifd->data_type==0x0c) printf("(DOUBLET [8 byte double])\n");
    else
  { printf("(Not Known)\n"); }

  printf("              Length: %d\n",tiff_ifd->length);
  printf("        Value/Offset: %d\n",tiff_ifd->value);

  if (tiff_ifd->data_type==0x02 && tiff_ifd->length>1)
  {
    marker=ftell(in);
    fseek(in,tiff_ifd->value,SEEK_SET);

    printf("\n\"");
    for (t=0; t<tiff_ifd->length; t++)
    {
      ch=getc(in);
      printf("%c",ch);
    }
    printf("\"\n");

    fseek(in,marker,SEEK_SET);
  }

  printf("\n");
#endif

  return 0;
}

int read_color_map(FILE *in, struct tiff_info_t *tiff_info, struct tiff_ifd_t *tiff_ifd)
{
int marker;
int l,r,g,b,ptr;

  marker=ftell(in);
  fseek(in,tiff_ifd->value,SEEK_SET);

  tiff_info->max_colors=tiff_ifd->length/((tiff_info->bits_per_sample/8)*3);
  tiff_info->color_map=malloc(tiff_info->max_colors*sizeof(int));

  l=0;
  ptr=0;

  if (tiff_info->bits_per_sample==8)
  {
    while (l<tiff_ifd->length)
    {
      r=getc(in);
      g=getc(in);
      b=getc(in);

      tiff_info->color_map[ptr++]=(r<<16)+(g<<8)+b;
      l=l+3;
      if (ptr>=tiff_info->max_colors) break;
    }
  }
    else
  if (tiff_info->bits_per_sample==16)
  {
    while (l<tiff_ifd->length)
    {
      if (tiff_info->endian==1)
      {
        r=read_word_b(in)>>8;
        g=read_word_b(in)>>8;
        b=read_word_b(in)>>8;
      }
        else
      {
        r=read_word(in)>>8;
        g=read_word(in)>>8;
        b=read_word(in)>>8;
      }

      tiff_info->color_map[ptr++]=(r<<16)+(g<<8)+b;
      if (ptr>=tiff_info->max_colors) break;
      l=l+6;
    }
  }

  fseek(in,marker,SEEK_SET);

  return 0;
}

unsigned int *read_tiff(FILE *in, int *width, int *height)
{
unsigned int *picture;
struct tiff_info_t tiff_info;
struct tiff_header_t tiff_header;
struct tiff_ifd_t tiff_ifd;
int ifd_count;
int min_sample_value,max_sample_value,photometric;
int t,c;

  picture=0;
  *width=0;
  *height=0;
  ifd_count=0;
  min_sample_value=0x00;
  max_sample_value=0xff;
  photometric=0;

  memset(&tiff_info,0,sizeof(tiff_info));
  tiff_info.strip_count=1;

  read_tiff_header(in,&tiff_header);

  if (tiff_header.byte_order[0]=='M')
  { tiff_info.endian=1; }
    else
  { tiff_info.endian=0; }

  fseek(in,tiff_header.pointer_to_ifd,SEEK_SET);

  if (tiff_info.endian==1)
  { ifd_count=read_word_b(in); }
    else
  { ifd_count=read_word(in); }

  printf("           IFD Count: %d\n",ifd_count);
  printf("\n");

  for (t=0; t<ifd_count; t++)
  {
    read_tiff_ifd(in,&tiff_ifd,tiff_info.endian);

    if (tiff_ifd.tag_type==0x100)
    { tiff_info.width=tiff_ifd.value; }
      else
    if (tiff_ifd.tag_type==0x101)
    { tiff_info.height=tiff_ifd.value; }
      else
    if (tiff_ifd.tag_type==0x103)
    { tiff_info.compression=tiff_ifd.value; }
      else
    if (tiff_ifd.tag_type==0x102)
    { tiff_info.bits_per_sample=tiff_ifd.value; }
      else
    if (tiff_ifd.tag_type==0x111)
    {
      tiff_info.strip_offset=tiff_ifd.value;
      tiff_info.strip_count=tiff_ifd.length;
    }
      else
    if (tiff_ifd.tag_type==0x118)
    { min_sample_value=tiff_ifd.value; }
      else
    if (tiff_ifd.tag_type==0x119)
    { max_sample_value=tiff_ifd.value; }
      else
    if (tiff_ifd.tag_type==0x106)
    { photometric=tiff_ifd.value; }
      else
    if (tiff_ifd.tag_type==0x140)
    { read_color_map(in,&tiff_info,&tiff_ifd); }
  }

#ifdef DEBUG
printf("      width: %d\n",tiff_info.width);
printf("     height: %d\n",tiff_info.height);
printf("compression: %d ",tiff_info.compression);

if (tiff_info.compression==1) printf("(uncompressed)\n");
  else
if (tiff_info.compression==2) printf("(Modified Huffman RLE)\n");
  else
if (tiff_info.compression==3) printf("(CCITT Group 3)\n");
  else
if (tiff_info.compression==4) printf("(CCITT Group 4)\n");
  else
if (tiff_info.compression==5) printf("(LZW)\n");
  else
if (tiff_info.compression==6) printf("(JPEG [old style])\n");
  else
if (tiff_info.compression==7) printf("(JPEG [new style])\n");
  else
if (tiff_info.compression==8) printf("(DEFLATE)\n");
  else
if (tiff_info.compression==32771) printf("(Uncompressed word boundary)\n");
  else
if (tiff_info.compression==32773) printf("(Packed Bits)\n");
  else
{ printf("(unknown)\n"); }

if (tiff_info.max_colors!=0)
{ printf(" num colors: %d\n",tiff_info.max_colors); }
printf("     strips: %d\n",tiff_info.strip_count);
printf("\n");
#endif

  picture=malloc(tiff_info.width*tiff_info.height*sizeof(int));
  memset(picture,0,tiff_info.width*tiff_info.height);
  *width=tiff_info.width;
  *height=tiff_info.height;

  if (tiff_info.rows_per_strip==0) tiff_info.rows_per_strip=tiff_info.height;

  if (tiff_info.color_map==0)
  {
    tiff_info.color_map=malloc(2*sizeof(int));
    tiff_info.max_colors=2;


    if (photometric==0)
    {
      if (min_sample_value==0)
      {
        tiff_info.color_map[0]=0x000000;
        tiff_info.color_map[1]=0xffffff;
      }
        else
      {
        tiff_info.color_map[0]=0xffffff;
        tiff_info.color_map[1]=0x000000;
      }
    }
      else
    {
      if (min_sample_value==0)
      {
        tiff_info.color_map[0]=0xffffff;
        tiff_info.color_map[1]=0x000000;
      }
        else
      {
        tiff_info.color_map[0]=0x000000;
        tiff_info.color_map[1]=0xffffff;
      }
    }
  }

  for (c=0; c<tiff_info.strip_count; c++)
  {
    tiff_info.y_offset=c*tiff_info.rows_per_strip;


    if (tiff_info.strip_count==1)
    { t=tiff_info.strip_offset; }
      else
    { t=read_long_offset(in,tiff_info.strip_offset+(c*4)); }

    fseek(in,t,SEEK_SET);
/* printf("%d 0x%06x\n",t,t); */

#ifdef DEBUG
printf("strip %d offset: %d\n",c+1,t);
#endif

    if (tiff_info.compression==1)
    { read_uncomp_tiff(in,picture,&tiff_info); }
      else
    if (tiff_info.compression==2)
    { read_huffman_tiff(in,picture,&tiff_info); }
      else
    if (tiff_info.compression==3)
    {
      /* tiff_info.color_map[1]=0x000000; */
      /* tiff_info.color_map[2]=0xffffff; */
      read_ccitt_tiff(in,picture,&tiff_info,3); 
    }
      else
    if (tiff_info.compression==4)
    {
      /* tiff_info.color_map[1]=0x000000; */
      /* tiff_info.color_map[2]=0xffffff; */
      read_ccitt_tiff(in,picture,&tiff_info,4); 
    }
      else
    if (tiff_info.compression==5)
    { lzw_decompress(in,picture,&tiff_info); }
      else
    if (tiff_info.compression==32773)
    { read_packed_tiff(in,picture,&tiff_info); }
      else
    { printf("This compression method is not supported at this time.\n"); }
  }

  if (tiff_info.color_map!=0) free(tiff_info.color_map);

  return picture;
}




